<?php

namespace SuperRhino\Utils\Cache;

use SuperRhino\Utils\Redis\Redis;
use SuperRhino\Utils\Time\InteractsWithTime;

class RedisCacheAdapter implements DriverCacheAdapterInterface {

    use InteractsWithTime;

    /**`
     * @var \Redis
     */
    protected $redis;
    /**
     * @var PackerInterface
     */
    protected $packer;
    /**
     * @var string
     */
    protected $prefix;

    public function __construct(string $prefix = 'c:') {
        $this->packer = new JsonPacker();
        $this->redis = Redis::getRedis(Redis::POOL_NAME_CACHE);
        $this->prefix = $prefix;
    }

    protected function getCacheKey(string $key) {
        return $this->prefix . $key;
    }

    private function pack($data, ?float $delta = null): string {
        if ($delta !== null) {
            return $this->packer->pack(['#!' => $data, '@!' => $delta]);
        }
        return $this->packer->pack($data);
    }

    private function unpack(string $data) {
        $pack = $this->packer->unpack($data);
        if (isset($pack['#!']) && isset($pack['@!'])) {
            return [$pack['#!'], $pack['@!']];
        }
        return [$pack, null];
    }

    public function get($key, $default = null) {
        $res = $this->redis->get($this->getCacheKey($key));
        if ($res === false) {
            return $default;
        }

        return $this->unpack($res)[0];
    }

    public function getCallback($key, callable $callback, $ttl = 10 * 60, ?float $beta = null) {
        if (0 > $beta = $beta ?? 1.0) {
            throw new InvalidArgumentException(sprintf('Argument "$beta" provided to "%s::getCallback()" must be a positive number, %f given.', static::class, $beta));
        }

        $data = $this->redis->get($this->getCacheKey($key));
        if ($data === false || $beta === INF) {
            return $this->callbackSaveCache($key, $callback, $ttl);
        }

        [$data, $delta] = $this->unpack($data);
        $time = microtime(true);
        $expiry = (max($this->redis->pttl($this->getCacheKey($key)), 0) / 1000) + $time;
        if ($time - max($delta, 10) * $beta * log(random_int(1, \PHP_INT_MAX) / \PHP_INT_MAX) >= $expiry) {
            return $this->callbackSaveCache($key, $callback, $ttl);
        }
        return $data;
    }

    private function callbackSaveCache($key, callable $callback, $ttl) {
        $start = intval(microtime(true) * 1000);
        $data = $callback();
        if ($data !== null) {
            $seconds = $this->secondsUntil($ttl);
            $pack = $this->pack($data, intval(microtime(true) * 1000) - $start);
            $this->redis->set($this->getCacheKey($key), $pack, $seconds > 0 ? $seconds : null);
            return $data;
        }
        return null;
    }

    public function fetch(string $key, $default = null): array {
        $res = $this->redis->get($this->getCacheKey($key));
        if ($res === false) {
            return [false, $default];
        }

        return [true, $this->unpack($res)[0]];
    }

    public function set($key, $value, $ttl = 10 * 60) {
        $seconds = $this->secondsUntil($ttl);
        $res = $this->pack($value);
        return $this->redis->set($this->getCacheKey($key), $res, $seconds > 0 ? $seconds : null);
    }

    public function delete($key) {
        return (bool)$this->redis->del($this->getCacheKey($key));
    }

    public function clear() {
        return $this->clearPrefix('');
    }

    public function getMultiple($keys, $default = null) {
        $cacheKeys = array_map(function ($key) {
            return $this->getCacheKey($key);
        }, $keys);

        $values = $this->redis->mget($cacheKeys);
        $result = [];
        foreach ($keys as $i => $key) {
            $result[$key] = $values[$i] === false ? $default : $this->unpack($values[$i][0]);
        }

        return $result;
    }

    public function setMultiple($values, $ttl = null) {
        if (!is_array($values)) {
            throw new InvalidArgumentException('The values is invalid!');
        }

        $cacheKeys = [];
        foreach ($values as $key => $value) {
            $cacheKeys[$this->getCacheKey($key)] = $this->pack($value);
        }

        $seconds = $this->secondsUntil($ttl);
        if ($seconds > 0) {
            foreach ($cacheKeys as $key => $value) {
                $this->redis->set($key, $value, $seconds);
            }

            return true;
        }

        return $this->redis->mset($cacheKeys);
    }

    public function deleteMultiple($keys) {
        $cacheKeys = array_map(function ($key) {
            return $this->getCacheKey($key);
        }, $keys);

        return (bool)$this->redis->del(...$cacheKeys);
    }

    public function has($key) {
        return (bool)$this->redis->exists($this->getCacheKey($key));
    }

    public function clearPrefix(string $prefix): bool {
        $iterator = null;
        $key = $prefix . '*';
        while (true) {
            $keys = $this->redis->scan($iterator, $this->getCacheKey($key), 10000);
            if (!empty($keys)) {
                $this->redis->del(...$keys);
            }

            if (empty($iterator)) {
                break;
            }
        }

        return true;
    }

    public function addKey(string $collector, string $key): bool {
        return (bool)$this->redis->sAdd($this->getCacheKey($collector), $key);
    }

    public function keys(string $collector): array {
        return $this->redis->sMembers($this->getCacheKey($collector)) ?? [];
    }

    public function delKey(string $collector, ...$key): bool {
        return (bool)$this->redis->sRem($this->getCacheKey($collector), ...$key);
    }
}